Generate sample prompts to trigger the generated flows, by fitting the flow name and description:

The prompts should:
- Be in imperative mood
- Be specific per flow
- Be under 72 characters

Generated flows:
{{ flows }}

Chose most important flows.
Generate only one prompt per flow
Return max 3 prompts ( at least one ) in a bullet point list: e.g.
- *Why is my internet slow?*
- *How do i reboot my router?*
