# Security limits for backup archive extraction
MAX_BACKUP_SIZE = 100 * 1024 * 1024  # 100MB
MAX_ARCHIVE_FILE_SIZE = 100 * 1024 * 1024  # 100MB
MAX_ARCHIVE_TOTAL_SIZE = 500 * 1024 * 1024  # 500MB
MAX_ARCHIVE_FILES = 10000
