SECRET_MANAGER_ENV_NAME = "SECRET_MANAGER"
SUPPORTED_SECRET_MANAGERS = ["vault"]

SECRET_MANAGER_KEY = "secrets_manager"
SECRETS_PATH = "secrets_path"
TRANSIT_MOUNT_POINT_LABEL = "transit_mount_point"
NAMESPACE_KEY_LABEL = "namespace"

SECRET_MANAGER_PREFIX = "secrets_manager"

SOURCE_KEY_LABEL = "source"
SECRET_KEY_LABEL = "secret_key"
TRANSIT_KEY_FOR_ENCRYPTION_LABEL = "transit_key"
TRACKER_STORE_ENDPOINT_TYPE = "tracker_store"

USERNAME_LABEL = "username"
PASSWORD_LABEL = "password"

VAULT_URL_ENV_NAME = "VAULT_URL"
VAULT_TOKEN_ENV_NAME = "VAULT_TOKEN"
VAULT_RASA_SECRETS_PATH_ENV_NAME = "VAULT_RASA_SECRETS_PATH"
VAULT_TRANSIT_MOUNT_POINT_ENV_NAME = "VAULT_TRANSIT_MOUNT_POINT"
VAULT_NAMESPACE_ENV_NAME = "VAULT_NAMESPACE"
VAULT_DEFAULT_RASA_SECRETS_PATH = "rasa-secrets"
VAULT_SECRET_MANAGER_NAME = "vault"
VAULT_MOUNT_POINT_ENV_NAME = "VAULT_MOUNT_POINT"


VAULT_ENDPOINT_URL_LABEL = "url"
VAULT_ENDPOINT_TOKEN_LABEL = "token"
VAULT_ENDPOINT_SECRETS_PATH_LABEL = "secrets_path"
VAULT_ENDPOINT_TRANSIT_MOUNT_POINT_LABEL = "transit_mount_point"
VAULT_ENDPOINT_NAMESPACE_LABEL = "namespace"
VAULT_ENDPOINT_MOUNT_POINT_LABEL = "mount_point"

VAULT_MOUNT_POINT_DEFAULT_VALUE = "secret"
