"""Constants for MCP server."""

# MCP Tool Names - Centralized constants for tool names used throughout the codebase
MCP_TOOL_SEARCH_DOCS = "search_docs"
MCP_TOOL_READ_PROJECT_FILES = "read_project_files"
MCP_TOOL_GET_PROJECT_FILE = "get_project_file"
MCP_TOOL_LIST_PROJECT_FILES = "list_project_files"
MCP_TOOL_WRITE_PROJECT_FILE = "write_project_file"
MCP_TOOL_UPDATE_MULTIPLE_FILES = "update_multiple_files"
MCP_TOOL_VALIDATE_PROJECT = "validate_project"
MCP_TOOL_TRAIN_MODEL = "train_model"
MCP_TOOL_GET_ASSISTANT_LOGS = "get_assistant_logs"
MCP_TOOL_TALK_TO_ASSISTANT = "talk_to_assistant"
